@@ -7,6 +7,7 @@ package netutil
 
 import (
 	"net"
+	"net/netip"
 
 	glnetutil "github.com/AdguardTeam/golibs/netutil"
 	"golang.org/x/exp/slices"
@@ -49,3 +50,30 @@ func SortIPAddrs(addrs []net.IPAddr, preferIPv6 bool) {
 		return a.Less(b)
 	})
 }
+
+// SortNetIPAddrs sorts addrs in accordance with the protocol preferences.
+// Invalid addresses are sorted near the end.  Zones are ignored.
+func SortNetIPAddrs(addrs []netip.Addr, preferIPv6 bool) {
+	l := len(addrs)
+	if l <= 1 {
+		return
+	}
+
+	slices.SortStableFunc(addrs, func(addrA, addrB netip.Addr) (sortsBefore bool) {
+		if !addrA.IsValid() {
+			return false
+		} else if !addrB.IsValid() {
+			return true
+		}
+
+		if aIs4, bIs4 := addrA.Is4(), addrB.Is4(); aIs4 != bIs4 {
+			if aIs4 {
+				return !preferIPv6
+			}
+
+			return preferIPv6
+		}
+
+		return addrA.Less(addrB)
+	})
+}
